
library(mvtnorm)
library(Matrix)


ProgressBar = function(maxn,ind){
  if(ind==1){
    cat("|1%--------------------50%--------------------100%|\n")
    cat("|")
    return()
  }
  if(maxn==ind){cat("|\n");return()}
  if(floor(50*ind/maxn) != floor(50*(ind-1)/maxn)){cat("|")}
}

d_all = function(Y,str,stp,norm="Frob",usecor=FALSE){
  p=nrow(Y)
  n=ncol(Y)
  ret_v = rep(0,stp-str+1)
  if(!usecor){
    S1 = t(Y[1:str,])%*%Y[1:str,]*(1/str)
    S2 = (t(Y[(str+1):p,])%*%Y[(str+1):p,])*(1/(p-str))
  }
  for(i in 1:(stp-str+1)){
    if(usecor){
      S1=cor(Y[1:(str+i),])
      S2=cor(Y[(str+i+1):p,])
    }else{
      changemat=matrix(t(Y[str+i,]),ncol=1)%*%matrix(Y[str+i,],nrow=1)
      S1=(S1*(str+i-1)+changemat)/(str+i)
      S2=(S2*(p-str-i+1)-changemat)/(p-str-i)      
    }
    difmat=S1-S2
    if(norm=="Frob"){
      ret_v[i]=sum(diag((difmat)%*%difmat))      
    }else if(norm=="Max"){
      diag(difmat)=0
      ret_v[i]=max(abs(difmat))
    }
  }
  return(ret_v)
}


d_allMult = function(Y,str,stp,norm="Frob",usecor=FALSE){
  p=nrow(Y[[1]])
  n=ncol(Y[[1]])
  ret_v = rep(0,stp-str+1)
  S1_v = list()
  S2_v = list()
  if(!usecor){
    for(ii in 1:length(Y)){
      ele=Y[[ii]]
      S1_v[[ii]] = t(ele[1:str,])%*%ele[1:str,]*(1/str)
      S2_v[[ii]] = (t(ele[(str+1):p,])%*%ele[(str+1):p,])*(1/(p-str))    
    }
  }
  for(i in 1:(stp-str+1)){
    S1 = matrix(0,nrow=n,ncol=n)
    S2 = matrix(0,nrow=n,ncol=n)
    for(ii in 1:length(Y)){
      ele = Y[[ii]]
      if(usecor){
        S1_v[[ii]]=cor(ele[1:(str+i),])
        S2_v[[ii]]=cor(ele[(str+i+1):p,])
      }else{
        changemat=matrix(t(ele[str+i,]),ncol=1)%*%matrix(ele[str+i,],nrow=1)
        S1_v[[ii]]=(S1_v[[ii]]*(str+i-1)+changemat)/(str+i)
        S2_v[[ii]]=(S2_v[[ii]]*(p-str-i+1)-changemat)/(p-str-i)              
      }
      S1 = S1+S1_v[[ii]]/length(Y)
      S2 = S2+S2_v[[ii]]/length(Y)
    }
    difmat=S1-S2
    if(norm=="Frob"){
      ret_v[i]=sum(diag((difmat)%*%difmat))      
    }else if(norm=="Max"){
      diag(difmat)=0
      ret_v[i]=max(abs(difmat))
    }
  }
  return(ret_v)
}






# d = function(Y,k){
#   p=nrow(Y)
#   n=ncol(Y)
#   difmat=t(Y[1:k,])%*%Y[1:k,]*(1/k)-(t(Y[(k+1):p,])%*%Y[(k+1):p,])*(1/(p-k))
#   sum(diag((difmat)%*%difmat))
# }

SieveBootstrapYMult = function(Ys,B=1000,stepsback=2,offset,display=TRUE,norm="Frob",usecor=FALSE){
  n=ncol(Ys[[1]])
  p=nrow(Ys[[1]])
  eps_mat_ls = list()
  exp_mat_ls = list()
  for(ii in 1:length(Ys)){
    ele=Ys[[ii]]
    eps_mat = matrix(,ncol=n,nrow=p-stepsback)
    exp_mat = matrix(,ncol=n,nrow=p-stepsback)
    rho_mat = matrix(,nrow=n,ncol=stepsback) # last column is one step back.
    INDs=rep(1:stepsback,p-stepsback)+ rep(0:(p-stepsback-1),each=stepsback)
    for(i in 1:n){
      A = matrix(ele[INDs,i],nrow=p-stepsback,ncol=stepsback,byrow=T)    
      c = ele[(stepsback+1):p,i]
      rho_mat[i,]=solve(t(A)%*%A)%*%t(A)%*%c
      exp_mat[,i]=A%*%rho_mat[i,]
      eps_mat[,i]=c-exp_mat[,i]
    }
    eps_mat_ls[[ii]]=eps_mat
    exp_mat_ls[[ii]]=exp_mat
  }
  dboot = matrix(,nrow=p,ncol=B)
  for(b in 1:B){
    INDboot=sample(1:(p-stepsback),replace=TRUE)
    Ysboot_ls = list()
    for(ii in 1:length(Ys)){
      ele=Ys[[ii]]
      exp_mat = exp_mat_ls[[ii]]
      eps_mat = eps_mat_ls[[ii]]
      Ysboot_ls[[ii]]=rbind(ele[1:stepsback,],exp_mat + eps_mat[INDboot,])
    }
    dboot[(offset):(p-offset-1),b]=d_allMult(Ysboot_ls,offset,p-offset-1,norm,usecor)
    if(display){ProgressBar(B,b)}
  }
  return(dboot[(offset):(p-offset-1),])
}


SieveBootstrapY = function(Ys,B=1000,stepsback=2,offset,display=TRUE,norm="Frob",usecor=FALSE){
  n=ncol(Ys)
  p=nrow(Ys)
  eps_mat = matrix(,ncol=n,nrow=p-stepsback)
  exp_mat = matrix(,ncol=n,nrow=p-stepsback)
  rho_mat = matrix(,nrow=n,ncol=stepsback) # last column is one step back.
  INDs=rep(1:stepsback,p-stepsback)+ rep(0:(p-stepsback-1),each=stepsback)
  for(i in 1:n){
    A = matrix(Ys[INDs,i],nrow=p-stepsback,ncol=stepsback,byrow=T)    
    c = Ys[(stepsback+1):p,i]
    rho_mat[i,]=solve(t(A)%*%A)%*%t(A)%*%c
    exp_mat[,i]=A%*%rho_mat[i,]
    eps_mat[,i]=c-exp_mat[,i]
  }
  dboot = matrix(,nrow=p,ncol=B)
  for(b in 1:B){
    INDboot=sample(1:(p-stepsback),replace=TRUE)
    Ysboot=rbind(Ys[1:stepsback,],exp_mat + eps_mat[INDboot,])
    dboot[(offset):(p-offset-1),b]=d_all(Ysboot,offset,p-offset-1,norm,usecor)
    if(display){ProgressBar(B,b)}
  }
  return(dboot[(offset):(p-offset-1),])
}

CPdetectionMult = function(Y,B=500,offset=NA,stepsback=2,norm="Frob",usecor=FALSE,display=TRUE){
  n=ncol(Y[[1]])
  p=nrow(Y[[1]])
  N=length(Y)
  if(is.na(offset)){offset=max(2,ceiling(n/N))}
  if(2*offset>=p-2){stop("Offset too large. Not enough data.")}
  Ys=list()
  for(i in 1:length(Y)){
    Ys[[i]]=scale(Y[[i]])/sqrt((p-1)/p)  
  }
  dboot = SieveBootstrapYMult(Ys,B=B,stepsback=stepsback,offset=offset,display=display,norm=norm,usecor=usecor)
  b_mean = apply(dboot,1,mean)
  b_sd = apply(dboot,1,sd)
  dboot = (dboot-b_mean)/b_sd
  maxd_boot=apply(dboot,2,max)
  dreal=d_allMult(Ys,offset,p-offset-1,norm=norm,usecor)
  dreal = (dreal-b_mean)/b_sd
  realmax = max(dreal)
  INDmax = which(dreal==realmax)  
  pval = length(which(maxd_boot>realmax))/B
  return(list("pvalue"=pval,"CPind"=INDmax+offset,"dscaled"=dreal))
  
}
  

CPdetection = function(Y,B=500,offset=NA,stepsback=1,norm="Frob",usecor=FALSE,display=TRUE){
  if(norm!="Frob" && norm!="Max"){stop("You must use 'Frob' or 'Max' as the norm")}
  if(is.list(Y)){
    return(CPdetectionMult(Y,B,offset,stepsback,norm,usecor,display))
  }
  n=ncol(Y)
  p=nrow(Y)
  if(is.na(offset)){offset=n}
  if(2*offset>=p-2){stop("Offset too large. Not enough data.")}
  Ys = scale(Y)/sqrt((p-1)/p)  
  dboot = SieveBootstrapY(Ys,B=B,stepsback=stepsback,offset=offset,display=display,norm=norm,usecor=usecor)
  b_mean = apply(dboot,1,mean)
  b_sd = apply(dboot,1,sd)
  dboot = (dboot-b_mean)/b_sd
  maxd_boot=apply(dboot,2,max)
  dreal=d_all(Ys,offset,p-offset-1,norm=norm,usecor=usecor)
  dreal = (dreal-b_mean)/b_sd
  realmax = max(dreal)
  INDmax = which(dreal==realmax)  
  pval = length(which(maxd_boot>realmax))/B
  return(list("pvalue"=pval,"CPind"=INDmax+offset,"dscaled"=dreal))
}


n = 10
p = 50
rho=0.5
Sigma1 = diag(n)*(1-rho)+matrix(rho,nrow=n,ncol=n) 
Sigma2t = cbind(diag(floor(n/2))*(1-rho)+matrix(rho,nrow=floor(n/2),ncol=floor(n/2)),matrix(0,nrow=floor(n/2),ncol=n-floor(n/2)))
Sigma2b = cbind(matrix(0,nrow=floor(n/2),ncol=n-floor(n/2)),diag(floor(n/2))*(1-rho)+matrix(rho,nrow=floor(n/2),ncol=floor(n/2)))
Sigma2 = rbind(Sigma2t,Sigma2b)

Y = rbind(rmvnorm(floor(p/2),mean=rep(0,n),sigma=Sigma1),rmvnorm(floor(p/2),mean=rep(0,n),sigma=Sigma2))
Sigma0=cor(Y)
pvals = rep(0,500)
for(j in 1:500){
  Y0=rmvnorm(p,mean=rep(0,n),sigma=Sigma0)
  Y02 = rmvnorm(p,mean=rep(0,n),sigma=Sigma0)
  pvals[j]=CPdetectionMult(list(Y0,Y02),B=200,offset=n/2,display=FALSE)$pvalue
  cat(j)
}
hist(pvals)
# 
# B=1000
# d_v = matrix(NA,nrow=B,ncol=p)
# offset=5
# for(b in 1:B){
#   Ysim=rmvnorm(p,mean=rep(0,n),sigma=Sigma0)
#   for(k in (1+offset):(p-offset-1)){
#     d_v[b,k]=d(Ysim,k)
#   }
#   ProgressBar(B,b)
# }
# 
# 
# 
# 
